from typing import List, Optional, Union
from uuid import UUID

from alitra import Frame, Orientation, Pose, Position
from pydantic import BaseModel, Field


class StepResponse(BaseModel):
    id: UUID
    type: str


class TaskResponse(BaseModel):
    id: UUID
    tag_id: Optional[str]
    steps: List[StepResponse]


class StartMissionResponse(BaseModel):
    id: Union[UUID, int, str, None]
    tasks: List[TaskResponse]


class ControlMissionResponse(BaseModel):
    mission_id: Union[UUID, int, str, None]
    mission_status: str
    task_id: Union[UUID, int, str, None]
    task_status: str
    step_id: Union[UUID, int, str, None]
    step_status: str


# We need to specify our own position/orientation/pose classes that do not contain
# the "Frame" class because of a bug in generating the OpenAPI specification:
# https://github.com/tiangolo/fastapi/issues/1505 This does not happen if all the
# classes are 'BaseModel' classes, but the alitra models are 'dataclasses', hence the
# conversion being done.
class InputOrientation(BaseModel):
    x: float
    y: float
    z: float
    w: float
    frame_name: str = Field(default="robot")

    def to_alitra_orientation(self) -> Orientation:
        return Orientation(
            x=self.x,
            y=self.y,
            z=self.z,
            w=self.w,
            frame=Frame(self.frame_name),
        )


class InputPosition(BaseModel):
    x: float
    y: float
    z: float
    frame_name: str = Field(default="robot")

    def to_alitra_position(self) -> Position:
        return Position(
            x=self.x,
            y=self.y,
            z=self.z,
            frame=Frame(self.frame_name),
        )


class InputPose(BaseModel):
    position: InputPosition
    orientation: InputOrientation
    frame_name: str = Field(default="robot")

    def to_alitra_pose(self) -> Pose:
        return Pose(
            position=self.position.to_alitra_position(),
            orientation=self.orientation.to_alitra_orientation(),
            frame=Frame(self.frame_name),
        )



import logging
from http import HTTPStatus
from logging import Logger
from typing import Union

import click
import uvicorn
from fastapi import FastAPI, Security
from fastapi.middleware.cors import CORSMiddleware
from fastapi.routing import APIRouter
from injector import inject
from pydantic import AnyHttpUrl

from isar.apis.models.models import StartMissionResponse
from isar.apis.schedule.scheduling_controller import SchedulingController
from isar.apis.security.authentication import Authenticator
from isar.config.settings import settings


class API:
    @inject
    def __init__(
        self,
        authenticator: Authenticator,
        scheduling_controller: SchedulingController,
        port: int = settings.API_PORT,
    ) -> None:

        self.authenticator: Authenticator = authenticator
        self.scheduling_controller: SchedulingController = scheduling_controller
        self.host: str = "0.0.0.0"  # Locking uvicorn to use 0.0.0.0
        self.port: int = port

        self.logger: Logger = logging.getLogger("api")

        self.app: FastAPI = self._create_app()

    def get_app(self) -> FastAPI:
        return self.app

    def run_app(self) -> None:
        uvicorn.run(
            self.app,
            port=self.port,
            host=self.host,
            reload=False,
            log_config=None,
        )

    def _create_app(self) -> FastAPI:
        tags_metadata = [
            {
                "name": "Scheduler",
                "description": "Mission functionality",
            }
        ]
        app = FastAPI(
            openapi_tags=tags_metadata,
            on_startup=[self.authenticator.load_config, self._log_startup_message],
            swagger_ui_oauth2_redirect_url="/oauth2-redirect",
            swagger_ui_init_oauth={
                "usePkceWithAuthorizationCodeGrant": True,
                "clientId": self.authenticator.openapi_client_id,
            },
        )

        if self.authenticator.should_authenticate():
            backend_cors_origins: list[Union[str, AnyHttpUrl]] = [
                f"http://{self.host}:{self.port}"
            ]

            app.add_middleware(
                CORSMiddleware,
                allow_origins=[str(origin) for origin in backend_cors_origins],
                allow_credentials=True,
                allow_methods=["*"],
                allow_headers=["*"],
            )

        app.include_router(router=self._create_scheduler_router())

        return app

    def _create_scheduler_router(self) -> APIRouter:

        router: APIRouter = APIRouter(tags=["Scheduler"])

        authentication_dependency: Security = Security(self.authenticator.get_scheme())

        router.add_api_route(
            "/schedule/start-mission/{id}",
            self.scheduling_controller.start_mission_by_id,
            methods=["POST"],
            dependencies=[authentication_dependency],
            summary="Start a mission with id='id' from the current mission planner",
            responses={
                HTTPStatus.OK.value: {
                    "description": "Mission succesfully started",
                    "model": StartMissionResponse,
                },
                HTTPStatus.NOT_FOUND.value: {
                    "description": "Not found - Mission not found",
                },
                HTTPStatus.CONFLICT.value: {
                    "description": "Conflict - Invalid command in the current state",
                },
                HTTPStatus.REQUEST_TIMEOUT.value: {
                    "description": "Timeout - Could not contact state machine",
                },
                HTTPStatus.INTERNAL_SERVER_ERROR.value: {
                    "description": "Internal Server Error - Current state of state machine unknown",
                },
            },
        )
        router.add_api_route(
            "/schedule/start-mission",
            self.scheduling_controller.start_mission,
            methods=["POST"],
            dependencies=[authentication_dependency],
            summary="Start the mission provided in JSON format",
            responses={
                HTTPStatus.OK.value: {
                    "description": "Mission succesfully started",
                    "model": StartMissionResponse,
                },
                HTTPStatus.UNPROCESSABLE_ENTITY.value: {
                    "description": "Invalid body - The JSON is incorrect",
                },
                HTTPStatus.CONFLICT.value: {
                    "description": "Conflict - Invalid command in the current state",
                },
                HTTPStatus.REQUEST_TIMEOUT.value: {
                    "description": "Timeout - Could not contact state machine",
                },
                HTTPStatus.INTERNAL_SERVER_ERROR.value: {
                    "description": "Internal Server Error - Current state of state machine unknown",
                },
            },
        )
        router.add_api_route(
            "/schedule/stop-mission",
            self.scheduling_controller.stop_mission,
            methods=["POST"],
            dependencies=[authentication_dependency],
            summary="Stop the current mission",
            responses={
                HTTPStatus.OK.value: {
                    "description": "Mission succesfully stopped",
                },
                HTTPStatus.CONFLICT.value: {
                    "description": "Conflict - Invalid command in the current state",
                },
                HTTPStatus.REQUEST_TIMEOUT.value: {
                    "description": "Timeout - Could not contact state machine",
                },
                HTTPStatus.INTERNAL_SERVER_ERROR.value: {
                    "description": "Internal Server Error - Current state of state machine unknown",
                },
            },
        )
        router.add_api_route(
            "/schedule/pause-mission",
            self.scheduling_controller.pause_mission,
            methods=["POST"],
            dependencies=[authentication_dependency],
            summary="Pause the current mission",
            responses={
                HTTPStatus.OK.value: {
                    "description": "Mission succesfully paused",
                },
                HTTPStatus.REQUEST_TIMEOUT.value: {
                    "description": "Timeout - Could not contact state machine",
                },
                HTTPStatus.CONFLICT.value: {
                    "description": "Conflict - Invalid command in the current state",
                },
                HTTPStatus.INTERNAL_SERVER_ERROR.value: {
                    "description": "Internal Server Error - Current state of state machine unknown",
                },
            },
        )
        router.add_api_route(
            "/schedule/resume-mission",
            self.scheduling_controller.resume_mission,
            methods=["POST"],
            dependencies=[authentication_dependency],
            summary="Resume the currently paused mission - if any",
            responses={
                HTTPStatus.OK.value: {
                    "description": "Mission succesfully resumed",
                },
                HTTPStatus.REQUEST_TIMEOUT.value: {
                    "description": "Timeout - Could not contact state machine",
                },
                HTTPStatus.CONFLICT.value: {
                    "description": "Conflict - Invalid command in the current state",
                },
                HTTPStatus.INTERNAL_SERVER_ERROR.value: {
                    "description": "Internal Server Error - Current state of state machine unknown",
                },
            },
        )
        router.add_api_route(
            "/schedule/drive-to",
            self.scheduling_controller.drive_to,
            methods=["POST"],
            dependencies=[authentication_dependency],
            summary="Drive to the provided pose",
            responses={
                HTTPStatus.OK.value: {
                    "description": "Drive to succesfully started",
                },
                HTTPStatus.REQUEST_TIMEOUT.value: {
                    "description": "Timeout - Could not contact state machine",
                },
                HTTPStatus.CONFLICT.value: {
                    "description": "Conflict - Invalid command in the current state",
                },
                HTTPStatus.INTERNAL_SERVER_ERROR.value: {
                    "description": "Internal Server Error - Current state of state machine unknown",
                },
            },
        )

        return router

    def _log_startup_message(self) -> None:
        address_format = "%s://%s:%d/docs"
        message = f"Uvicorn running on {address_format} (Press CTRL+C to quit)"
        protocol = "http"
        color_message = (
            "Uvicorn running on "
            + click.style(address_format, bold=True)
            + " (Press CTRL+C to quit)"
        )

        self.logger.info(
            message,
            protocol,
            self.host,
            self.port,
            extra={"color_message": color_message},
        )



from http import HTTPStatus
import json

import pytest
from isar.apis.models.models import InputOrientation, InputPosition
from fastapi.testclient import TestClient
from fastapi.encoders import jsonable_encoder
from isar.apis.security.authentication import Authenticator
from isar.apis.models import InputPose
from isar.mission_planner.local_planner import LocalPlanner
from isar.mission_planner.mission_planner_interface import MissionPlannerError
from isar.models.communication.queues.queue_timeout_error import QueueTimeoutError
from isar.services.utilities.scheduling_utilities import SchedulingUtilities
from isar.state_machine.states_enum import States
from tests.mocks.mission_definition import MockMissionDefinition

mock_orientation: InputOrientation = InputOrientation(
    x=0, y=0, z=0, w=1, frame_name="robot"
)
mock_position: InputPosition = InputPosition(x=1, y=1, z=1, frame_name="robot")
mock_pose: InputPose = InputPose(
    orientation=mock_orientation, position=mock_position, frame_name="robot"
)


def mock_check_queue(was_mission_started, state_at_request):
    mock_return_1 = was_mission_started, state_at_request
    if was_mission_started or state_at_request != "idle":
        mock_return_2 = False
    else:
        mock_return_2 = True

    return [mock_return_1, mock_return_2]


class TestSchedulerRoutes:
    @pytest.mark.parametrize(
        "mission_id, mock_get_mission, expected_exception,"
        "mock_get_state, mock_start_mission_side_effect, expected_status_code",
        [
            (
                12345,
                MockMissionDefinition.default_mission,
                MissionPlannerError,
                States.Idle,
                None,
                HTTPStatus.INTERNAL_SERVER_ERROR,
            ),
            (
                1,
                MockMissionDefinition.default_mission,
                None,
                States.Idle,
                QueueTimeoutError,
                HTTPStatus.REQUEST_TIMEOUT,
            ),
            (
                1,
                MockMissionDefinition.default_mission,
                None,
                States.Monitor,
                True,
                HTTPStatus.CONFLICT,
            ),
            (
                1,
                MockMissionDefinition.default_mission,
                None,
                States.Idle,
                QueueTimeoutError,
                HTTPStatus.REQUEST_TIMEOUT,
            ),
            (
                1,
                MockMissionDefinition.default_mission,
                None,
                States.Idle,
                None,
                HTTPStatus.OK,
            ),
        ],
    )
    def test_start_mission_by_id(
        self,
        client,
        access_token,
        mocker,
        mission_id,
        mock_get_mission,
        expected_exception,
        mock_get_state,
        mock_start_mission_side_effect,
        expected_status_code,
    ):
        mocker.patch.object(
            LocalPlanner,
            "get_mission",
            return_value=mock_get_mission,
            side_effect=expected_exception,
        )

        mocker.patch.object(
            SchedulingUtilities,
            "get_state",
            return_value=mock_get_state,
        )

        mocker.patch.object(
            SchedulingUtilities,
            "start_mission",
            side_effect=mock_start_mission_side_effect,
        )

        response = client.post(
            f"schedule/start-mission/{mission_id}",
            headers={"Authorization": "Bearer {}".format(access_token)},
        )

        assert response.status_code == expected_status_code

    @pytest.mark.parametrize(
        "state, mock_return, expected_status_code",
        [
            (
                States.Monitor,
                QueueTimeoutError(),
                HTTPStatus.REQUEST_TIMEOUT,
            ),
            (
                States.InitiateStep,
                None,
                HTTPStatus.OK,
            ),
            (
                States.Idle,
                None,
                HTTPStatus.CONFLICT,
            ),
        ],
    )
    def test_stop_mission(
        self,
        client,
        access_token,
        mocker,
        state,
        mock_return,
        expected_status_code,
    ):
        mocker.patch.object(SchedulingUtilities, "get_state", return_value=state)
        mocker.patch.object(
            SchedulingUtilities, "stop_mission", side_effect=mock_return
        )

        response = client.post(
            "/schedule/stop-mission",
            headers={"Authorization": "Bearer {}".format(access_token)},
        )

        assert response.status_code == expected_status_code

    @pytest.mark.parametrize(
        "mock_start_mission, mock_get_state,expected_status_code,target_pose",
        [
            (
                None,
                States.Idle,
                HTTPStatus.OK,
                mock_pose,
            ),
            (
                QueueTimeoutError,
                States.Idle,
                HTTPStatus.REQUEST_TIMEOUT,
                mock_pose,
            ),
            (
                None,
                States.Monitor,
                HTTPStatus.CONFLICT,
                mock_pose,
            ),
            (
                None,
                States.Idle,
                HTTPStatus.UNPROCESSABLE_ENTITY,
                InputPosition(
                    x=1,
                    y=1,
                    z=1,
                    frame_name="robot",
                ),
            ),
        ],
    )
    def test_schedule_drive_to(
        self,
        client: TestClient,
        access_token,
        mock_start_mission,
        mock_get_state,
        mocker,
        expected_status_code,
        target_pose,
    ):
        mocker.patch.object(
            SchedulingUtilities, "get_state", return_value=mock_get_state
        )

        mocker.patch.object(
            SchedulingUtilities, "start_mission", side_effect=mock_start_mission
        )
        mocker.patch.object(Authenticator, "should_authenticate", return_value=False)

        data: str = json.dumps(jsonable_encoder(target_pose))

        response = client.post(
            url=f"/schedule/drive-to",
            headers={"Authorization": "Bearer {}".format(access_token)},
            data=data,
        )

        assert response.status_code == expected_status_code



from setuptools import find_packages, setup

setup(
    name="isar",
    description="Integration and Supervisory control of Autonomous Robots",
    long_description=open("README.md").read(),
    long_description_content_type="text/markdown",
    author="Equinor ASA",
    author_email="fg_robots_dev@equinor.com",
    url="https://github.com/equinor/isar",
    packages=find_packages(where="src"),
    package_dir={"": "src"},
    classifiers=[
        "Environment :: Other Environment",
        "Intended Audience :: Developers",
        "Intended Audience :: Science/Research",
        "Programming Language :: Python",
        "Topic :: Scientific/Engineering",
        "Topic :: Scientific/Engineering :: Physics",
        "Topic :: Software Development :: Libraries",
    ],
    include_package_data=True,
    install_requires=[
        "alitra>=1.1.0",
        "azure-identity",
        "azure-keyvault-secrets",
        "azure-storage-blob",
        "backoff",
        "click",
        "dacite",
        "fastapi-azure-auth",
        "fastapi",
        "injector",
        "numpy",
        "paho-mqtt",
        "pydantic",
        "PyJWT",
        "python-dotenv",
        "PyYAML",
        "requests-toolbelt",
        "requests",
        "transitions",
        "uvicorn",
    ],
    extras_require={
        "dev": [
            "black",
            "flake8",
            "mypy",
            "myst-parser",
            "pytest-dotenv",
            "pytest-mock",
            "pytest-xdist",
            "pytest",
            "requests-mock",
            "sphinx",
            "pre-commit",
        ]
    },
    python_requires=">=3.9",
)



import logging
import time
from typing import TYPE_CHECKING, Optional

from transitions import State

from isar.models.communication.message import StartMissionMessage

if TYPE_CHECKING:
    from isar.state_machine.state_machine import StateMachine


class Idle(State):
    def __init__(self, state_machine: "StateMachine"):
        super().__init__(name="idle", on_enter=self.start, on_exit=self.stop)
        self.state_machine: "StateMachine" = state_machine
        self.logger = logging.getLogger("state_machine")

    def start(self):
        self.state_machine.update_state()
        self._run()

    def stop(self):
        pass

    def _run(self):
        while True:
            start_mission: Optional[
                StartMissionMessage
            ] = self.state_machine.should_start_mission()
            if start_mission:
                self.state_machine.start_mission(
                    mission=start_mission.mission,
                    initial_pose=start_mission.initial_pose,
                )
                transition = self.state_machine.mission_started
                break
            time.sleep(self.state_machine.sleep_time)

        transition()



import json
import logging
import queue
import time
from collections import deque
from datetime import datetime
from typing import Deque, List, Optional

from alitra import Pose
from injector import Injector, inject
from transitions import Machine
from transitions.core import State
from isar.apis.models.models import ControlMissionResponse

from isar.config.settings import settings
from isar.mission_planner.task_selector_interface import (
    TaskSelectorInterface,
    TaskSelectorStop,
)
from isar.models.communication.message import StartMissionMessage
from isar.models.communication.queues.queues import Queues
from isar.models.mission import Mission, Task
from isar.models.mission.status import MissionStatus, TaskStatus
from isar.state_machine.states import (
    Idle,
    Initialize,
    InitiateStep,
    Monitor,
    Off,
    Paused,
    StopStep,
)
from isar.state_machine.states_enum import States
from robot_interface.models.initialize.initialize_params import InitializeParams
from robot_interface.models.mission import StepStatus
from robot_interface.models.mission.step import Step
from robot_interface.robot_interface import RobotInterface
from robot_interface.telemetry.mqtt_client import MqttClientInterface
from robot_interface.utilities.json_service import EnhancedJSONEncoder


class StateMachine(object):
    """Handles state transitions for supervisory robot control."""

    @inject
    def __init__(
        self,
        queues: Queues,
        robot: RobotInterface,
        mqtt_publisher: MqttClientInterface,
        task_selector: TaskSelectorInterface,
        sleep_time: float = settings.FSM_SLEEP_TIME,
        stop_robot_attempts_limit: int = settings.STOP_ROBOT_ATTEMPTS_LIMIT,
        transitions_log_length: int = settings.STATE_TRANSITIONS_LOG_LENGTH,
    ):
        """Initializes the state machine.

        Parameters
        ----------
        queues : Queues
            Queues used for API communication.
        robot : RobotInterface
            Instance of robot interface.
        mqtt_publisher : MqttClientInterface
            Instance of MQTT client interface which has a publish function
        sleep_time : float
            Time to sleep in between state machine iterations.
        stop_robot_attempts_limit : int
            Maximum attempts to stop the robot when stop command is received
        transitions_log_length : int
            Length of state transition log list.

        """
        self.logger = logging.getLogger("state_machine")

        self.queues: Queues = queues
        self.robot: RobotInterface = robot
        self.mqtt_publisher: Optional[MqttClientInterface] = mqtt_publisher
        self.task_selector: TaskSelectorInterface = task_selector

        # List of states
        self.stop_step_state: State = StopStep(self)
        self.paused_state: State = Paused(self)
        self.idle_state: State = Idle(self)
        self.initialize_state: State = Initialize(self)
        self.monitor_state: State = Monitor(self)
        self.initiate_step_state: State = InitiateStep(self)
        self.off_state: State = Off(self)

        self.states: List[State] = [
            self.off_state,
            self.idle_state,
            self.initialize_state,
            self.initiate_step_state,
            self.monitor_state,
            self.stop_step_state,
            self.paused_state,
        ]

        self.machine = Machine(
            self,
            states=self.states,
            initial="off",
            queued=True,
        )

        self.machine.add_transitions(
            [
                {
                    "trigger": "step_initiated",
                    "source": self.initiate_step_state,
                    "dest": self.monitor_state,
                    "before": self._step_initiated,
                },
                {
                    "trigger": "pause",
                    "source": [self.initiate_step_state, self.monitor_state],
                    "dest": self.stop_step_state,
                    "before": self._pause,
                },
                {
                    "trigger": "stop",
                    "source": [self.initiate_step_state, self.monitor_state],
                    "dest": self.stop_step_state,
                    "before": self._stop,
                },
                {
                    "trigger": "mission_finished",
                    "source": [
                        self.initiate_step_state,
                    ],
                    "dest": self.idle_state,
                    "before": self._mission_finished,
                },
                {
                    "trigger": "mission_started",
                    "source": self.idle_state,
                    "dest": self.initialize_state,
                    "before": self._mission_started,
                },
                {
                    "trigger": "initialization_successful",
                    "source": self.initialize_state,
                    "dest": self.initiate_step_state,
                    "before": self._initialization_successful,
                },
                {
                    "trigger": "initialization_failed",
                    "source": self.initialize_state,
                    "dest": self.idle_state,
                    "before": self._initialization_failed,
                },
                {
                    "trigger": "resume",
                    "source": self.paused_state,
                    "dest": self.initiate_step_state,
                    "before": self._resume,
                },
                {
                    "trigger": "step_finished",
                    "source": self.monitor_state,
                    "dest": self.initiate_step_state,
                    "before": self._step_finished,
                },
                {
                    "trigger": "mission_paused",
                    "source": self.stop_step_state,
                    "dest": self.paused_state,
                    "before": self._mission_paused,
                },
                {
                    "trigger": "step_infeasible",
                    "source": self.initiate_step_state,
                    "dest": self.initiate_step_state,
                    "before": self._step_infeasible,
                },
                {
                    "trigger": "initiate_step_failed",
                    "source": self.initiate_step_state,
                    "dest": self.idle_state,
                    "before": self._initiate_step_failed,
                },
                {
                    "trigger": "mission_stopped",
                    "source": [self.stop_step_state, self.paused_state],
                    "dest": self.idle_state,
                    "before": self._mission_stopped,
                },
            ]
        )

        self.stop_robot_attempts_limit: int = stop_robot_attempts_limit
        self.sleep_time: float = sleep_time

        self.stopped: bool = False
        self.current_mission: Optional[Mission] = None
        self.current_task: Optional[Task] = None
        self.current_step: Optional[Step] = None
        self.initial_pose: Optional[Pose] = None

        self.current_state: State = States(self.state)  # type: ignore

        self.predefined_mission_id: Optional[int] = None

        self.transitions_log_length: int = transitions_log_length
        self.transitions_list: Deque[States] = deque([], self.transitions_log_length)

    #################################################################################
    # Transition Callbacks
    def _initialization_successful(self) -> None:
        self.queues.start_mission.output.put(True)
        self.logger.info(
            f"Initialization successful. Starting new mission: {self.current_mission.id}"
        )
        self.log_step_overview(mission=self.current_mission)

        # This is a workaround to enable the Flotilla repository to write the mission to
        # its database before the publishing from ISAR starts. This is not a permanent
        # solution and should be further addressed in the following issue.
        # https://github.com/equinor/flotilla/issues/226
        time.sleep(2)

        self.current_mission.status = MissionStatus.InProgress
        self.publish_mission_status()
        self.current_task = self.task_selector.next_task()
        self.current_task.status = TaskStatus.InProgress
        self.publish_task_status()
        self.update_current_step()

    def _initialization_failed(self) -> None:
        self.queues.start_mission.output.put(False)
        self._finalize()

    def _step_initiated(self) -> None:
        self.current_step.status = StepStatus.InProgress
        self.publish_step_status()
        self.logger.info(
            f"Successfully initiated "
            f"{type(self.current_step).__name__} "
            f"step: {str(self.current_step.id)[:8]}"
        )

    def _pause(self) -> None:
        return

    def _resume(self) -> None:
        self.logger.info(f"Resuming mission: {self.current_mission.id}")
        self.current_mission.status = MissionStatus.InProgress
        self.current_task.status = TaskStatus.InProgress
        self.publish_mission_status()
        self.publish_task_status()

        resume_mission_response: ControlMissionResponse = (
            self._make_control_mission_response()
        )
        self.queues.resume_mission.output.put(resume_mission_response)

        self.current_task.reset_task()
        self.update_current_step()

    def _mission_finished(self) -> None:
        self.current_mission.status = MissionStatus.Completed
        self._finalize()

    def _mission_started(self) -> None:
        return

    def _step_finished(self) -> None:
        self.publish_step_status()
        self.update_current_task()
        self.update_current_step()

    def _mission_paused(self) -> None:
        self.logger.info(f"Pausing mission: {self.current_mission.id}")
        self.current_mission.status = MissionStatus.Paused
        self.current_task.status = TaskStatus.Paused
        self.current_step.status = StepStatus.NotStarted

        paused_mission_response: ControlMissionResponse = (
            self._make_control_mission_response()
        )
        self.queues.pause_mission.output.put(paused_mission_response)

        self.publish_mission_status()
        self.publish_task_status()
        self.publish_step_status()

    def _stop(self) -> None:
        self.stopped = True

    def _initiate_step_failed(self) -> None:
        self.current_step.status = StepStatus.Failed
        self.current_mission.status = MissionStatus.Failed
        self.publish_step_status()
        self._finalize()

    def _step_infeasible(self) -> None:
        self.current_step.status = StepStatus.Failed
        self.publish_step_status()
        self.update_current_task()
        self.update_current_step()

    def _mission_stopped(self) -> None:
        self.current_mission.status = MissionStatus.Cancelled
        for task in self.current_mission.tasks:
            for step in task.steps:
                if step.status in [StepStatus.NotStarted, StepStatus.InProgress]:
                    step.status = StepStatus.Cancelled
            if task.status in [
                TaskStatus.NotStarted,
                TaskStatus.InProgress,
                TaskStatus.Paused,
            ]:
                task.status = TaskStatus.Cancelled

        stopped_mission_response: ControlMissionResponse = (
            self._make_control_mission_response()
        )
        self.queues.stop_mission.output.put(stopped_mission_response)

        self.publish_task_status()
        self.publish_step_status()
        self._finalize()

    #################################################################################

    def _finalize(self) -> None:
        self.publish_mission_status()
        self.log_step_overview(mission=self.current_mission)
        state_transitions: str = ", ".join(
            [
                f"\n  {transition}" if (i + 1) % 10 == 0 else f"{transition}"
                for i, transition in enumerate(list(self.transitions_list))
            ]
        )
        self.logger.info(f"State transitions:\n  {state_transitions}")
        self.reset_state_machine()

    def begin(self):
        """Starts the state machine.

        Transitions into idle state.

        """
        self.to_idle()

    def update_current_task(self):
        if self.current_task.is_finished():
            self.publish_task_status()
            try:
                self.current_task = self.task_selector.next_task()
                self.current_task.status = TaskStatus.InProgress
                self.publish_task_status()
            except TaskSelectorStop:
                # Indicates that all tasks are finished
                self.current_task = None

    def update_current_step(self):
        if self.current_task:
            self.current_step = self.current_task.next_step()

    def update_state(self):
        """Updates the current state of the state machine."""
        self.current_state = States(self.state)
        self.send_state_status()
        self._log_state_transition(self.current_state)
        self.logger.info(f"State: {self.current_state}")
        self.publish_state()

    def reset_state_machine(self) -> None:
        self.stopped = False
        self.current_step = None
        self.current_task = None
        self.current_mission = None

    def start_mission(self, mission: Mission, initial_pose: Pose):
        """Starts a scheduled mission."""
        self.current_mission = mission
        self.initial_pose = initial_pose

        self.task_selector.initialize(tasks=self.current_mission.tasks)

    def get_initialize_params(self):
        return InitializeParams(initial_pose=self.initial_pose)

    def should_start_mission(self) -> Optional[StartMissionMessage]:
        try:
            return self.queues.start_mission.input.get(block=False)
        except queue.Empty:
            return None

    def should_stop_mission(self) -> bool:
        try:
            return self.queues.stop_mission.input.get(block=False)
        except queue.Empty:
            return False

    def should_pause_mission(self) -> bool:
        try:
            return self.queues.pause_mission.input.get(block=False)
        except queue.Empty:
            return False

    def should_resume_mission(self) -> bool:
        try:
            return self.queues.resume_mission.input.get(block=False)
        except queue.Empty:
            return False

    def send_state_status(self):
        self.queues.state.update(self.current_state)

    def publish_mission_status(self) -> None:
        if not self.mqtt_publisher:
            return
        payload: str = json.dumps(
            {
                "robot_id": settings.ROBOT_ID,
                "mission_id": self.current_mission.id if self.current_mission else None,
                "status": self.current_mission.status if self.current_mission else None,
                "timestamp": datetime.utcnow(),
            },
            cls=EnhancedJSONEncoder,
        )

        self.mqtt_publisher.publish(
            topic=settings.TOPIC_ISAR_MISSION,
            payload=payload,
            retain=False,
        )

    def publish_task_status(self) -> None:
        """Publishes the current task status to the MQTT Broker"""
        if not self.mqtt_publisher:
            return
        payload: str = json.dumps(
            {
                "robot_id": settings.ROBOT_ID,
                "mission_id": self.current_mission.id if self.current_mission else None,
                "task_id": self.current_task.id if self.current_task else None,
                "status": self.current_task.status if self.current_task else None,
                "timestamp": datetime.utcnow(),
            },
            cls=EnhancedJSONEncoder,
        )

        self.mqtt_publisher.publish(
            topic=settings.TOPIC_ISAR_TASK,
            payload=payload,
            retain=False,
        )

    def publish_step_status(self) -> None:
        """Publishes the current step status to the MQTT Broker"""
        if not self.mqtt_publisher:
            return
        payload: str = json.dumps(
            {
                "robot_id": settings.ROBOT_ID,
                "mission_id": self.current_mission.id if self.current_mission else None,
                "task_id": self.current_task.id if self.current_task else None,
                "step_id": self.current_step.id if self.current_step else None,
                "status": self.current_step.status if self.current_step else None,
                "timestamp": datetime.utcnow(),
            },
            cls=EnhancedJSONEncoder,
        )

        self.mqtt_publisher.publish(
            topic=settings.TOPIC_ISAR_STEP,
            payload=payload,
            retain=False,
        )

    def publish_state(self) -> None:
        if not self.mqtt_publisher:
            return
        payload: str = json.dumps(
            {
                "robot_id": settings.ROBOT_ID,
                "state": self.current_state,
                "timestamp": datetime.utcnow(),
            },
            cls=EnhancedJSONEncoder,
        )

        self.mqtt_publisher.publish(
            topic=settings.TOPIC_ISAR_STATE,
            payload=payload,
            retain=True,
        )

    def _log_state_transition(self, next_state):
        """Logs all state transitions that are not self-transitions."""
        self.transitions_list.append(next_state)

    def log_step_overview(self, mission: Mission):
        """Log an overview of the steps in a mission"""
        log_statements: List[str] = []
        for task in mission.tasks:
            log_statements.append(
                f"{type(task).__name__:<20} {str(task.id)[:8]:<32} -- {task.status}"
            )
            for j, step in enumerate(task.steps):
                log_statements.append(
                    f"{j:>3} {type(step).__name__:<20} {str(step.id)[:8]:<32} -- {step.status}"  # noqa: E501
                )

        log_statement: str = "\n".join(log_statements)

        self.logger.info(f"Mission overview:\n{log_statement}")

    def _make_control_mission_response(self) -> ControlMissionResponse:
        return ControlMissionResponse(
            mission_id=self.current_mission.id,
            mission_status=self.current_mission.status,
            task_id=self.current_task.id,
            task_status=self.current_task.status,
            step_id=self.current_step.id,
            step_status=self.current_step.status,
        )


def main(injector: Injector):
    """Starts a state machine instance."""
    state_machine = injector.get(StateMachine)
    state_machine.begin()



import logging
from pathlib import Path

from isar.config.settings import settings
from isar.models.mission_metadata.mission_metadata import MissionMetadata
from isar.storage.storage_interface import StorageException, StorageInterface
from isar.storage.utilities import construct_local_paths, construct_metadata_file
from robot_interface.models.inspection.inspection import Inspection


class LocalStorage(StorageInterface):
    def __init__(self):
        self.root_folder: Path = Path(settings.LOCAL_STORAGE_PATH)
        self.logger = logging.getLogger("uploader")

    def store(self, inspection: Inspection, metadata: MissionMetadata):
        local_path, local_metadata_path = construct_local_paths(
            inspection=inspection, metadata=metadata
        )

        absolute_path: Path = self.root_folder.joinpath(local_path)
        absolute_metadata_path: Path = self.root_folder.joinpath(local_metadata_path)

        absolute_path.parent.mkdir(parents=True, exist_ok=True)

        metadata_bytes: bytes = construct_metadata_file(
            inspection=inspection, metadata=metadata, filename=local_path.name
        )
        try:
            with open(absolute_path, "wb") as file, open(
                absolute_metadata_path, "wb"
            ) as metadata_file:
                file.write(inspection.data)
                metadata_file.write(metadata_bytes)
        except IOError as e:
            self.logger.warning(
                f"Failed open/write for one of the following files: \n"
                f"{absolute_path}\n{absolute_metadata_path}"
            )
            raise StorageException from e
        except Exception as e:
            self.logger.error(
                "An unexpected error occurred while writing to local storage"
            )
            raise StorageException from e



class RobotException(Exception):
    pass


class RobotCommunicationException(RobotException):
    pass


class RobotInfeasibleStepException(RobotException):
    pass


class RobotInvalidResponseException(RobotException):
    pass


class RobotMapException(RobotException):
    pass


class RobotInvalidTelemetryException(RobotException):
    pass
